# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.
"""Input fields for tasks."""

import abc
import enum
from collections.abc import Callable, Generator
from collections.abc import Collection as ABCCollection
from operator import attrgetter
from types import GenericAlias
from typing import (
    Any,
    ClassVar,
    NotRequired,
    Self,
    TypedDict,
    Unpack,
    cast,
    override,
)

from debusine.artifacts.models import ArtifactCategory
from debusine.client.models import RelationType
from debusine.tasks.executors import ExecutorImageCategory
from debusine.tasks.inputs import serializers
from debusine.tasks.inputs.resolver import FieldResolver
from debusine.tasks.models import (
    BaseDynamicTaskData,
    BaseTaskInputArtifact,
    ExtraExternalRepository,
    ExtraRepository,
    InputArtifactMultiple,
    InputArtifactSingle,
    InputCollectionSingle,
    LookupMultiple,
    LookupSingle,
)
from debusine.utils import extract_generic_type_arguments


class Stage(enum.IntEnum):
    """Sorted enumeration for stages in preparing tasks for scheduling."""

    #: Inputs needed to prepare tasks for configuration lookup
    PENDING = 1
    #: Inputs needed to prepare tasks for running
    RUNNING = 2


class TaskFieldContainer:
    """Class that can contain task fields."""

    #: Inputs defined in this task class
    inputs: dict[str, "TaskInput[Any]"] = {}

    @override
    def __init_subclass__(cls, **kwargs: Any) -> None:
        """Merge input field definitions from multiple parents."""
        super().__init_subclass__()
        inputs: dict[str, TaskInput[Any]] = {}
        for parent in reversed(cls.__mro__):
            if not issubclass(parent, TaskFieldContainer):
                continue
            inputs.update(parent.inputs)
        inputs.update(cls.inputs)
        cls.inputs = inputs

    def __init__(self) -> None:
        """Initialize a TaskFieldContainer."""
        super().__init__()
        #: If set, when fields of this container are resolved, fields of the
        #: child container are also resolved, using the resolver instantiated
        #: for the parent container
        self.child_task_field_container: TaskFieldContainer | None = None

    def iter_all_fields(
        self, stage: Stage | None = None
    ) -> Generator[tuple["TaskFieldContainer", "TaskInput[Any]"]]:
        """Enumerate the fields to resolve, and their container object."""
        if stage is None:
            stage = Stage.RUNNING

        for inp in self.inputs.values():
            if inp.stage > stage:
                continue
            yield self, inp
        if self.child_task_field_container is not None:
            yield from self.child_task_field_container.iter_all_fields(stage)

    def field_is_resolved(self, name: str) -> bool:
        """Return True if the field has been resolved."""
        return name in self.__dict__

    def save_field_values(self, dynamic_data: BaseDynamicTaskData) -> None:
        """
        Save field values in a BaseDynamicTaskData.

        Fields not defined in this container that are already saved are
        preserved: this allows multiple field containers to serialize to the
        same BaseDynamicTaskData, as long as their names do not conflict.
        """
        task_input_fields = dynamic_data.task_input_fields
        for cont, inp in self.iter_all_fields():
            task_input_fields[inp.name] = inp.serialize_value(
                getattr(cont, inp.name)
            )

        # We need to perform an assignment to task_input_fields and not just
        # modify it in place, otherwise the pydantic model will consider the
        # field unset and not serialize it
        dynamic_data.task_input_fields = task_input_fields

    def restore_field_values(self, dynamic_data: BaseDynamicTaskData) -> None:
        """
        Restore field values serialized in a BaseDynamicTaskData.

        Stored fields with names not defined in this container are ignored:
        this allows multiple field containers to serialize to the same
        BaseDynamicTaskData, as long as their names do not conflict.
        """
        values = dynamic_data.task_input_fields
        for cont, inp in self.iter_all_fields():
            if inp.name not in values:
                continue
            setattr(cont, inp.name, inp.deserialize_value(values[inp.name]))


class TaskInputKwargs(TypedDict):
    """Argument typing for TaskInput constructor."""

    stage: NotRequired[Stage]
    verbose_name: NotRequired[str]
    help_text: NotRequired[str]
    hidden: NotRequired[bool]
    internal: NotRequired[bool]


class TaskInput[ValueType](abc.ABC):
    """Representation for one or more input artifacts for a task."""

    name: str
    stage: Stage
    serializer: ClassVar[serializers.TaskInputSerializer[ValueType]]

    def get_resolver_method(
        self, resolver: FieldResolver
    ) -> Callable[[Self], ValueType]:
        """Override this to forward to a FieldResolver method."""
        raise NotImplementedError(
            "This input field does not forward resolution to the FieldResolver."
        )

    def resolve(self, resolver: FieldResolver) -> ValueType:
        """Resolve the input from task data."""
        method = self.get_resolver_method(resolver)
        return method(self)

    def get_artifact_ids(self, value: ValueType) -> list[int]:  # noqa: ARG002, U100
        """Return artifact IDs."""
        return []

    def __init__(
        self,
        stage: Stage = Stage.RUNNING,
        verbose_name: str | None = None,
        help_text: str | None = None,
        hidden: bool = False,
        internal: bool = False,
    ) -> None:
        """
        Declare an input field.

        :param stage: stage of task preparation when the field is needed
        :param verbose_name: label to use instead of the autogenerated one
        :param help_text: description of the field
        :param hidden: set to True to hide the field from the UI
        :param internal: show this widget as an internal field in the UI
        """
        self.stage = stage
        self.verbose_name = verbose_name
        self.help_text = help_text
        self.hidden = hidden
        self.internal = internal

    def __set_name__(self, owner: type[TaskFieldContainer], name: str) -> None:
        """Set the field name at class construction."""
        self.name = name
        # Avoid sharing the inputs definition across all subclasses of
        # BaseExternalTask
        if "inputs" not in owner.__dict__:
            owner.inputs = owner.inputs.copy()
        owner.inputs[name] = self

    def __get__(
        self,
        obj: TaskFieldContainer,
        _: type[TaskFieldContainer] | None = None,
    ) -> ValueType:
        """Look up this input in the task object."""
        from debusine.tasks import TaskConfigError

        # If __get__ is called, it means that self.name is not in obj.__dict__,
        # which only happens if the field value has not yet been resolved
        raise TaskConfigError(f"Cannot access unresolved input {self.name!r}")

    def get_value(self, field_container: TaskFieldContainer) -> ValueType:
        """Return the current value of this field in the container."""
        return cast(ValueType, getattr(field_container, self.name))

    def serialize_value(self, value: ValueType) -> Any:
        """Serialize a value for storing in dynamic task data."""
        return self.serializer.serialize(value)

    def deserialize_value(self, value: Any) -> ValueType:
        """Serialize a value for storing in dynamic task data."""
        return self.serializer.deserialize(value)


class DataFieldInputKwargs(TaskInputKwargs):
    """Argument typing for DataFieldInput constructor."""

    field: NotRequired[str]


class DataFieldInput[LookupType, ValueType](TaskInput[ValueType]):
    """Input field that resolves values based on a task data field."""

    field: str
    getter: Callable[[str], LookupType]
    # Actual lookup type, filled by __init_subclass__
    lookup_type: type[LookupType]

    @override
    def __init_subclass__(cls, **kwargs: Any) -> None:
        """Fill lookup_type from typing information."""
        super().__init_subclass__(**kwargs)
        [cls.lookup_type, _] = extract_generic_type_arguments(
            cls, DataFieldInput
        )
        # Given that we use lookup_type for isinstance checks, if it is a
        # generic alias we can strip the parameterization and only keep the
        # original type. This allows to type LookupType as ABCCollection[foo]
        # and have isinstance checks against ABCCollection
        if isinstance(cls.lookup_type, GenericAlias):
            cls.lookup_type = cls.lookup_type.__origin__

    def __init__(
        self, field: str | None = None, **kwargs: Unpack[TaskInputKwargs]
    ) -> None:
        """
        Declare an input defined by a task data field.

        :param field: dot-separated list of lookup names in task data. Defaults
          to the input member name
        """
        super().__init__(**kwargs)
        if field is not None:
            self._set_field_name(field)

    @override
    def __set_name__(self, owner: type[TaskFieldContainer], name: str) -> None:
        """Set the field name at class construction."""
        super().__set_name__(owner, name)
        if not hasattr(self, "field"):
            self._set_field_name(name)

    def _set_field_name(self, name: str) -> None:
        """Set the attribute name used to look up the value in task input."""
        self.field = name
        self.getter = attrgetter(name)


class DataListFieldInput[LookupType, ValueType](
    DataFieldInput[ABCCollection[LookupType], ValueType]
):
    """DataFieldInput that resolves to a list of type checked elements."""

    # Type of each list item, filled by __init_subclass__
    item_type: type[LookupType]

    @override
    def __init_subclass__(cls, **kwargs: Any) -> None:
        """Fill lookup_type from typing information."""
        super().__init_subclass__(**kwargs)
        [cls.item_type, _] = extract_generic_type_arguments(
            cls, DataListFieldInput
        )


class BaseArtifactInputKwargs(DataFieldInputKwargs):
    """Argument typing for BaseArtifactInput constructor."""

    categories: NotRequired[ABCCollection[ArtifactCategory]]


class BaseArtifactInput[LookupType, ValueType](
    DataFieldInput[LookupType, ValueType]
):
    """Look up artifacts enforcing their types."""

    categories: tuple[ArtifactCategory, ...] | None

    def __init__(
        self,
        *,
        categories: ABCCollection[ArtifactCategory] | None = None,
        **kwargs: Unpack[DataFieldInputKwargs],
    ) -> None:
        """
        Look up artifacts enforcing their type.

        :param field: dot-separated list of lookup names in task data
        :param categories: list of acceptable artifact categories
        """
        super().__init__(**kwargs)
        self.categories = tuple(categories) if categories is not None else None

    def check_category(
        self, info: BaseTaskInputArtifact, field_name: str | None = None
    ) -> None:
        """Check the category of a single input artifact."""
        # Prevent import loop
        from debusine.tasks import ensure_artifact_categories

        if self.categories is None:
            return

        # TODO: this should not happen, as we are checking the results of
        # database lookups which always have a category. However, info.category
        # can be None for compatibility during a transition time of merging
        # with InputArtifactMultiple, which needs to be able to deserialize
        # from only a list of artifact IDs
        assert info.category is not None

        ensure_artifact_categories(
            configuration_key=field_name or self.field,
            category=info.category,
            expected=self.categories,
        )


class ArtifactListInput[LookupType, ValueType](
    BaseArtifactInput[ABCCollection[LookupType], ValueType],
    DataListFieldInput[LookupType, ValueType],
):
    """Input taking a list of artifact lookups, enforcing their types."""


class SingleInput(BaseArtifactInput[LookupSingle, InputArtifactSingle]):
    """One single input artifact specified in a task data field."""

    serializer = serializers.PydanticSerializer(InputArtifactSingle)

    @override
    def resolve(self, resolver: FieldResolver) -> InputArtifactSingle:
        lookup = resolver.get_field(self)
        assert lookup is not None
        info = resolver.lookup_single_artifact(lookup, label=self.field)
        self.check_category(info)
        return info

    @override
    def get_artifact_ids(self, value: InputArtifactSingle) -> list[int]:
        return value.resolved_ids()


class OptionalSingleInput(
    BaseArtifactInput[LookupSingle, InputArtifactSingle | None]
):
    """One optional single input artifact specified in a task data field."""

    serializer = serializers.OptionalPydanticSerializer(InputArtifactSingle)

    @override
    def resolve(self, resolver: FieldResolver) -> InputArtifactSingle | None:
        lookup = resolver.get_field(self)
        if lookup is None:
            return None
        info = resolver.lookup_single_artifact(lookup, label=self.field)
        self.check_category(info)
        return info

    @override
    def get_artifact_ids(self, value: InputArtifactSingle | None) -> list[int]:
        if value is None:
            return []

        return value.resolved_ids()


class SingleInputList(
    ArtifactListInput[LookupSingle, list[InputArtifactSingle]]
):
    """A list of single input artifact lookups in a task data field."""

    serializer = serializers.PydanticListSerializer(InputArtifactSingle)

    @override
    def resolve(self, resolver: FieldResolver) -> list[InputArtifactSingle]:
        infos: list[InputArtifactSingle] = []
        for lookup in resolver.get_field_collection(self):
            assert isinstance(lookup, LookupSingle)
            info = resolver.lookup_single_artifact(lookup, label=self.field)
            self.check_category(info)
            infos.append(info)
        return infos

    @override
    def get_artifact_ids(self, value: list[InputArtifactSingle]) -> list[int]:
        return [
            artifact.artifact_id
            for artifact in value
            if artifact.artifact_id is not None
        ]


class MultiInput(BaseArtifactInput[LookupMultiple, InputArtifactMultiple]):
    """A list of multiple artifacts lookups specified in a task data field."""

    serializer = serializers.PydanticSerializer(InputArtifactMultiple)

    @override
    def resolve(self, resolver: FieldResolver) -> InputArtifactMultiple:
        lookup = resolver.get_field(self)
        assert lookup is not None
        result = resolver.lookup_multiple_artifacts(lookup, label=self.field)
        if self.categories is not None:
            for i, artifact in enumerate(result.artifacts):
                self.check_category(artifact, f"{self.field}[{i}]")
        return result

    @override
    def get_artifact_ids(self, value: InputArtifactMultiple) -> list[int]:
        return [
            artifact.artifact_id
            for artifact in value.artifacts
            if artifact.artifact_id is not None
        ]


class MultiInputList(
    ArtifactListInput[LookupMultiple, list[InputArtifactMultiple]]
):
    """A multiple artifacts lookup specified in a task data field."""

    serializer = serializers.MultiInputListSerializer()

    @override
    def resolve(self, resolver: FieldResolver) -> list[InputArtifactMultiple]:
        infos: list[InputArtifactMultiple] = []
        for lookup_idx, lookup in enumerate(
            resolver.get_field_collection(self)
        ):
            result = resolver.lookup_multiple_artifacts(
                lookup, label=self.field
            )
            if self.categories is not None:
                for result_idx, artifact in enumerate(result.artifacts):
                    self.check_category(
                        artifact, f"{self.field}[{lookup_idx}][{result_idx}]"
                    )
            infos.append(result)
        return infos

    @override
    def get_artifact_ids(self, value: list[InputArtifactMultiple]) -> list[int]:
        ids = []

        for input_artifact_multiple in value:
            ids += [
                artifact.artifact_id
                for artifact in input_artifact_multiple.artifacts
                if artifact.artifact_id is not None
            ]
        return ids


class UploadArtifactsInput(
    BaseArtifactInput[LookupMultiple, InputArtifactMultiple]
):
    """A multiple artifact lookup finding all related binaries in an upload."""

    serializer = serializers.PydanticSerializer(InputArtifactMultiple)

    @override
    def __init__(self, **kwargs: Unpack[BaseArtifactInputKwargs]) -> None:
        categories_with_upload = [ArtifactCategory.UPLOAD]
        categories_with_upload.extend(kwargs.get("categories") or ())
        kwargs["categories"] = categories_with_upload
        super().__init__(**kwargs)

    @override
    def resolve(self, resolver: FieldResolver) -> InputArtifactMultiple:
        lookup = resolver.get_field(self)
        assert lookup is not None

        # Incrementally build the result set
        result: list[BaseTaskInputArtifact] = []

        # Lookup the initial set of packages
        upload_ids: list[int] = []
        initial = resolver.lookup_multiple_artifacts(
            lookup, label=self.field
        ).artifacts
        for i, artifact in enumerate(initial):
            if artifact.category == ArtifactCategory.UPLOAD:
                # File uploads for later expansion
                assert artifact.artifact_id is not None
                upload_ids.append(artifact.artifact_id)
            else:
                self.check_category(artifact, f"{self.field}[{i}]")
                result.append(artifact)

        # Expand with binaries related to uploads
        if upload_ids:
            result.extend(
                resolver.find_related_artifacts(
                    upload_ids,
                    target_category=ArtifactCategory.BINARY_PACKAGE,
                    relation_type=RelationType.EXTENDS,
                    label=self.field,
                ).artifacts
            )

        return InputArtifactMultiple(
            lookup=lookup, label=self.field, artifacts=result
        )


class EnvironmentInputKwargs(BaseArtifactInputKwargs):
    """Argument typing for EnvironmentInput constructor."""

    image_category: NotRequired[ExecutorImageCategory | None]
    set_backend: NotRequired[bool]
    try_variant: NotRequired[bool]


class EnvironmentInput(BaseArtifactInput[LookupSingle, InputArtifactSingle]):
    """One environment artifact specified in a task data field."""

    serializer = serializers.PydanticSerializer(InputArtifactSingle)

    def __init__(
        self,
        *,
        image_category: ExecutorImageCategory | None = None,
        set_backend: bool = True,
        try_variant: bool = True,
        **kwargs: Unpack[BaseArtifactInputKwargs],
    ) -> None:
        """
        Look up artifacts enforcing their type.

        :param field: dot-separated list of lookup names in task data
        :param categories: list of acceptable artifact categories
        :param image_category: try to use an environment with this image
          category; defaults to the image category needed by the executor
          for `self.backend`
        :param set_backend: if True (default), try to use an environment
          matching `self.backend`
        :param try_variant: if True (default), try to use an environment
          whose variant is `self.name`, but fall back to looking up an
          environment without a variant if the first lookup fails
        """
        super().__init__(**kwargs)
        self.image_category = image_category
        self.set_backend = set_backend
        self.try_variant = try_variant

    @override
    def get_resolver_method(
        self, resolver: FieldResolver
    ) -> Callable[[Self], InputArtifactSingle]:
        return resolver.resolve_environmentinput

    @override
    def get_artifact_ids(self, value: InputArtifactSingle) -> list[int]:
        return value.resolved_ids()


class OptionalEnvironmentInput(
    BaseArtifactInput[LookupSingle, InputArtifactSingle | None]
):
    """One optional environment artifact specified in a task data field."""

    serializer = serializers.OptionalPydanticSerializer(InputArtifactSingle)

    @override
    def get_resolver_method(
        self, resolver: FieldResolver
    ) -> Callable[[Self], InputArtifactSingle | None]:
        return resolver.resolve_optionalenvironmentinput


class SuiteArchiveInput(
    DataFieldInput[
        LookupSingle, tuple[InputCollectionSingle, InputCollectionSingle | None]
    ]
):
    """Look up a suite and an archive based on a task data field."""

    serializer = serializers.SuiteArchiveInputSerializer()

    @override
    def get_resolver_method(
        self, resolver: FieldResolver
    ) -> Callable[
        [Self], tuple[InputCollectionSingle, InputCollectionSingle | None]
    ]:
        return resolver.resolve_suitearchiveinput


class ExtraRepositoriesInput(
    DataFieldInput[list[ExtraRepository], list[ExtraExternalRepository] | None]
):
    """Look up extra repositories."""

    serializer = serializers.ExtraRepositoriesSerializer()

    @override
    def get_resolver_method(
        self, resolver: FieldResolver
    ) -> Callable[[Self], list[ExtraExternalRepository] | None]:
        return resolver.resolve_extrarepositoriesinput


class DebusineFQDNInput(TaskInput[str]):
    """Resolve to the Debusine FQDN."""

    serializer = serializers.ScalarSerializer(str)

    @override
    def resolve(self, resolver: FieldResolver) -> str:
        return resolver.get_server_setting("DEBUSINE_FQDN")


class BuildArchitectureInput(DataFieldInput[str, str]):
    """
    Build architecture for the task.

    This currently looks up a value in the task data, and uses a fallback value
    if it is not set.
    """

    serializer = serializers.ScalarSerializer(str)

    @override
    def resolve(self, resolver: FieldResolver) -> str:
        if arch := resolver.get_field(self):
            return arch
        return resolver.get_fallback_architecture()


class UnsignedMultiInput(
    BaseArtifactInput[LookupMultiple, InputArtifactMultiple]
):
    """A multiple artifact lookup finding all related artifacts."""

    serializer = serializers.PydanticSerializer(InputArtifactMultiple)

    @override
    def resolve(self, resolver: FieldResolver) -> InputArtifactMultiple:
        lookup = resolver.get_field(self)
        assert lookup is not None

        result: list[BaseTaskInputArtifact] = []

        signed_artifacts: set[int] = set()

        for signed in resolver.lookup_multiple_artifacts(
            lookup, label=self.field
        ).artifacts:
            assert signed.artifact_id
            signed_artifacts.add(signed.artifact_id)

        signing_inputs: set[int] = set()

        for signing_input in resolver.find_related_artifacts(
            signed_artifacts,
            target_category=ArtifactCategory.SIGNING_INPUT,
            relation_type=RelationType.RELATES_TO,
            label=self.field,
        ).artifacts:
            assert signing_input.artifact_id
            result.append(signing_input)
            signing_inputs.add(signing_input.artifact_id)

        result += resolver.find_related_artifacts(
            signing_inputs,
            target_category=ArtifactCategory.BINARY_PACKAGE,
            relation_type=RelationType.RELATES_TO,
            label=self.field,
        ).artifacts

        def artifact_id(result: BaseTaskInputArtifact) -> int:
            assert result.artifact_id is not None
            return result.artifact_id

        result.sort(key=artifact_id)

        return InputArtifactMultiple(
            lookup=lookup, label=self.field, artifacts=result
        )



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""OpenMetrics statistics for Debusine."""

import datetime as dt
from collections.abc import Iterable, Sequence
from functools import cache
from itertools import product
from typing import Any, cast, override

from django.conf import settings
from django.db.models import (
    BooleanField,
    Count,
    Exists,
    ExpressionWrapper,
    F,
    IntegerField,
    Model,
    OuterRef,
    Q,
    Subquery,
    Sum,
)
from django.db.models.expressions import Combinable
from django.db.models.functions import Coalesce, Extract
from django.utils import timezone
from prometheus_client.core import (
    GaugeMetricFamily,
    HistogramMetricFamily,
    Metric,
    Timestamp,
)
from prometheus_client.registry import Collector

from debusine.artifacts.local_artifact import LocalArtifact
from debusine.artifacts.models import CollectionCategory, TaskTypes
from debusine.assets.models import AssetCategory, asset_categories
from debusine.db.models import (
    Artifact,
    Asset,
    Collection,
    FileStore,
    Group,
    Identity,
    Scope,
    Token,
    User,
    WorkRequest,
    Worker,
    WorkerPool,
    WorkerPoolStatistics,
    WorkflowTemplate,
    Workspace,
)
from debusine.db.tasks import DBTask
from debusine.tasks.executors import executor_backends
from debusine.tasks.models import WorkerType

BUCKET_DAYS = (1, 3, 7, 14, 30, 90, 365)


class ZeroGauge(GaugeMetricFamily):
    """A GaugeMetricFamily that can fill missing values with 0."""

    def __init__(
        self,
        name: str,
        documentation: str,
        labels: Sequence[str],
        **kwargs: Any,
    ) -> None:
        """Initialize a ZeroGauge."""
        super().__init__(name, documentation, labels=labels, **kwargs)
        self._seen: set[tuple[str, ...]] = set()

    @override
    def add_metric(
        self,
        labels: Sequence[str],
        value: float,
        timestamp: Timestamp | float | None = None,
    ) -> None:
        """Record a metric."""
        super().add_metric(labels, value, timestamp)
        self._seen.add(tuple(labels))

    def fill_zeros(self, *label_values: Iterable[str]) -> None:
        """Generate zero metrics for any label values we haven't seen."""
        for labels in product(*label_values):
            if labels not in self._seen:
                self.add_metric(labels, 0)
                self._seen.add(labels)


class DebusineCollector(Collector):
    """Collect Debusine's metrics from the database."""

    @override
    def collect(self) -> Iterable[Metric]:
        """Collect all our metrics."""
        self._scopes.cache_clear()
        self._worker_pools.cache_clear()

        yield self.count_by_category(
            Artifact, categories=LocalArtifact.artifact_categories()
        )
        yield self.count_assets()
        yield self.count_by_category(
            Collection, categories=list(CollectionCategory)
        )
        yield self.count_groups()
        yield self.count_tokens()
        yield self.count_user_identities()
        yield self.count_user_identities_activity()
        yield self.count_users()
        yield self.count_user_activity()
        yield self.count_work_requests()
        yield self.count_workers()
        yield self.count_worker_pool_runtime()
        yield self.count_workflow_templates()
        yield self.count_workspaces()
        yield from self.measure_file_stores()

    @cache
    def _scopes(self) -> Sequence[str]:
        """Return all known scope names."""
        return list(Scope.objects.values_list("name", flat=True))

    @cache
    def _worker_pools(self, include_null: bool = False) -> Sequence[str]:
        """Return all known worker pool names."""
        names = list(WorkerPool.objects.values_list("name", flat=True))
        if include_null:
            names.append("")
        return names

    def count_by_category(
        self, model: type[Model], categories: Sequence[str]
    ) -> ZeroGauge:
        """Collect the number of objects in model by category and scope."""
        gauge = ZeroGauge(
            str(model._meta.verbose_name_plural),
            f"Number of {model._meta.object_name}s",
            labels=("category", "scope"),
        )
        for stat in model._meta.base_manager.values(
            "category", "workspace__scope__name"
        ).annotate(count=Count("id")):
            gauge.add_metric(
                (stat["category"], stat["workspace__scope__name"] or ""),
                stat["count"],
            )
        gauge.fill_zeros(categories, self._scopes())
        return gauge

    def count_assets(self) -> Metric:
        """Count the number of assets by category and scope."""
        unscoped_categories = {AssetCategory.CLOUD_PROVIDER_ACCOUNT}
        scoped_categories = set(asset_categories()) - unscoped_categories
        metric = self.count_by_category(
            Asset, categories=tuple(scoped_categories)
        )
        metric.fill_zeros(tuple(unscoped_categories), ("",))
        return metric

    def measure_file_stores(self) -> Iterable[Metric]:
        """Measure the size of FileStores."""
        file_store_size = GaugeMetricFamily(
            "file_store_size",
            "total_size of FileStores",
            labels=("backend", "name"),
        )
        file_store_max_size = GaugeMetricFamily(
            "file_store_max_size",
            "max_size of FileStores",
            labels=("backend", "name"),
        )
        for (
            name,
            backend,
            current_total_size,
            max_size,
        ) in FileStore.objects.annotate_current_total_size().values_list(
            "name", "backend", "current_total_size", "max_size"
        ):
            file_store_size.add_metric((backend, name), current_total_size)
            if max_size:
                file_store_max_size.add_metric((backend, name), max_size)
        yield file_store_size
        yield file_store_max_size

    def count_groups(self) -> Metric:
        """Count the number of known groups."""
        gauge = ZeroGauge(
            "groups", "Number of Groups", labels=("ephemeral", "scope")
        )
        for stat in Group.objects.values("ephemeral", "scope__name").annotate(
            count=Count("id")
        ):
            gauge.add_metric(
                (str(int(stat["ephemeral"])), stat["scope__name"]),
                stat["count"],
            )
        gauge.fill_zeros(("0", "1"), self._scopes())
        return gauge

    def count_tokens(self) -> Metric:
        """Count the number of issued tokens."""
        # TODO: 2026-04-29: Remove user and worker labels after 6 months
        gauge = ZeroGauge(
            "tokens",
            "Number of Tokens",
            labels=("enabled", "user", "worker", "token_type"),
        )
        for stat in (
            Token.objects.values("enabled", "token_type")
            .annotate(
                user_token=ExpressionWrapper(
                    Q(token_type=Token.Types.USER), output_field=BooleanField()
                ),
                worker_token=ExpressionWrapper(
                    Q(token_type=Token.Types.WORKER),
                    output_field=BooleanField(),
                ),
            )
            .annotate(count=Count("id"))
        ):
            gauge.add_metric(
                (
                    str(int(stat["enabled"])),
                    str(int(stat["user_token"])),
                    str(int(stat["worker_token"])),
                    stat["token_type"],
                ),
                stat["count"],
            )

        for token_type in Token.Types:
            gauge.fill_zeros(
                ("0", "1"),
                (str(int(token_type == Token.Types.USER)),),
                (str(int(token_type == Token.Types.WORKER)),),
                (str(token_type),),
            )
        return gauge

    def count_users(self) -> Metric:
        """Count the number of known users."""
        gauge = ZeroGauge("users", "Number of Users", labels=("active",))
        for stat in (
            User.objects.exclude(username="_system")
            .values("is_active")
            .annotate(count=Count("id"))
        ):
            gauge.add_metric((str(int(stat["is_active"])),), stat["count"])
        gauge.fill_zeros(("0", "1"))
        return gauge

    def _histogram_buckets(self) -> dict[str, Count]:
        """Generate histogram buckets."""
        buckets = {}
        now = timezone.now()
        for days in BUCKET_DAYS:
            buckets[f'b{days}d'] = Count(
                "created_by",
                distinct=True,
                filter=Q(created_at__gte=now - dt.timedelta(days=days)),
            )
        buckets['binf'] = Count("created_by", distinct=True)
        return buckets

    def _histogram_count(
        self, histogram: HistogramMetricFamily, labels: tuple[str, ...]
    ) -> None:
        """Count user activity into histogram."""
        label_null_filters = {f"{label}__isnull": False for label in labels}

        for stat in (
            WorkRequest.objects.filter(
                task_type=TaskTypes.WORKFLOW, parent__isnull=True
            )
            .filter(**label_null_filters)
            .values(*labels)
            .annotate(**self._histogram_buckets())
        ):
            buckets = [(str(days), stat[f'b{days}d']) for days in BUCKET_DAYS]
            buckets.append(('+Inf', stat['binf']))
            histogram.add_metric(
                tuple(stat[label] for label in labels),
                buckets=buckets,
                sum_value=None,
            )

    def count_user_activity(self) -> Metric:
        """Count user activity."""
        histogram = HistogramMetricFamily(
            "user_activity",
            "Number of users who have created a workflow in the bucket",
            labels=("scope",),
        )
        self._histogram_count(histogram, labels=("workspace__scope__name",))
        return histogram

    def count_user_identities(self) -> Metric:
        """Count the number of known user identities."""
        gauge = ZeroGauge(
            "user_identities",
            "Number of User Identities",
            labels=("active", "issuer"),
        )
        for stat in (
            Identity.objects.filter(user__isnull=False)
            .values("user__is_active", "issuer")
            .annotate(count=Count("id"))
        ):
            gauge.add_metric(
                (str(int(stat["user__is_active"])), stat["issuer"]),
                stat["count"],
            )
        issuers = [
            provider.name
            for provider in getattr(settings, "SIGNON_PROVIDERS", ())
        ]
        gauge.fill_zeros(("0", "1"), issuers)
        return gauge

    def count_user_identities_activity(self) -> Metric:
        """Count user identity activity."""
        histogram = HistogramMetricFamily(
            "user_identities_activity",
            "Number of SSO Users who have created a workflow in the bucket",
            labels=("issuer", "scope"),
        )
        self._histogram_count(
            histogram,
            ("created_by__identities__issuer", "workspace__scope__name"),
        )
        return histogram

    def count_work_requests(self) -> Metric:
        """Count the number of known work requests."""
        gauge = ZeroGauge(
            "work_requests",
            "Number of WorkRequests",
            labels=(
                "task_type",
                "task_name",
                "scope",
                "status",
                "build_architecture",
                "backend",
            ),
        )
        seen_architectures = set()
        for stat in WorkRequest.objects.values(
            "task_type",
            "task_name",
            "workspace__scope__name",
            "status",
            "task_data__build_architecture",
            "task_data__backend",
        ).annotate(count=Count("id")):
            gauge.add_metric(
                (
                    stat["task_type"],
                    stat["task_name"],
                    stat["workspace__scope__name"],
                    stat["status"],
                    stat["task_data__build_architecture"] or "",
                    stat["task_data__backend"] or "",
                ),
                stat["count"],
            )
            if stat["task_data__build_architecture"]:
                seen_architectures.add(stat["task_data__build_architecture"])
        for task_type in TaskTypes:
            architectures: Iterable[str] = ("",)
            backends: Iterable[str] = ("",)
            if task_type == TaskTypes.WORKER:
                architectures = seen_architectures
                backends = executor_backends()
            gauge.fill_zeros(
                (task_type,),
                DBTask.task_names(task_type),
                self._scopes(),
                WorkRequest.Statuses,
                architectures,
                backends,
            )

        return gauge

    def count_workers(self) -> Metric:
        """Count the number of known workers."""
        name = "workers"
        description = "Number of Workers"
        labels = [
            "connected",
            "busy",
            "worker_type",
            "worker_pool",
            "native_architecture",
        ]
        # TODO: Drop system:host_architecture fallback once all workers have
        # been updated.
        workers = Worker.objects.annotate(
            static_native_architecture=Coalesce(
                "static_metadata__system:native_architecture",
                "static_metadata__system:host_architecture",
            ),
            dynamic_native_architecture=Coalesce(
                "dynamic_metadata__system:native_architecture",
                "dynamic_metadata__system:host_architecture",
            ),
        )
        architectures: set[str] = set()
        native_architectures: set[str] = set()
        for archs, native_arch in (
            workers.values_list(
                "static_metadata__system:architectures",
                "static_native_architecture",
            )
            .distinct(
                "static_metadata__system:architectures",
                "static_native_architecture",
            )
            .union(
                workers.values_list(
                    "dynamic_metadata__system:architectures",
                    "dynamic_native_architecture",
                ).distinct(
                    "dynamic_metadata__system:architectures",
                    "dynamic_native_architecture",
                )
            )
        ):
            if archs:
                for arch in archs:
                    architectures.add(arch)
            if native_arch:
                native_architectures.add(native_arch)

        labels += [f"architecture_{arch}" for arch in architectures]

        gauge = ZeroGauge(name, description, labels=labels)

        for stat in (
            workers.values(
                "worker_type",
                "worker_pool__name",
                "static_native_architecture",
                "dynamic_native_architecture",
                "static_metadata__system:architectures",
                "dynamic_metadata__system:architectures",
            )
            .annotate(
                connected=Exists(
                    Worker.objects.connected().filter(pk=OuterRef("pk"))
                ),
                busy=Exists(
                    WorkRequest.objects.filter(
                        status__in=(
                            WorkRequest.Statuses.RUNNING,
                            WorkRequest.Statuses.PENDING,
                        ),
                        worker=OuterRef("pk"),
                    )
                ),
            )
            .annotate(count=Count("id"))
        ):
            metric_labels = [
                str(int(stat["connected"])),
                str(int(stat["busy"])),
                stat["worker_type"],
                stat["worker_pool__name"] or "",
                stat["static_native_architecture"]
                or stat["dynamic_native_architecture"]
                or "",
            ]
            stat_archs = (
                stat["static_metadata__system:architectures"]
                or stat["dynamic_metadata__system:architectures"]
                or []
            )
            metric_labels += [
                str(int(arch in stat_archs)) for arch in architectures
            ]
            gauge.add_metric(metric_labels, stat["count"])
        for worker_type in WorkerType:
            label_values: list[Iterable[str]] = [
                # connected, busy, worker_type
                ("0", "1"),
                ("0", "1"),
                (worker_type,),
            ]
            if worker_type == WorkerType.EXTERNAL:
                # worker_pool, native_architecture, architecture_*
                label_values += [
                    self._worker_pools(include_null=True),
                    native_architectures,
                ]
                label_values += [("0", "1")] * len(architectures)
            else:
                label_values += [("",), ("",)]
                label_values += [("0",)] * len(architectures)
            gauge.fill_zeros(*label_values)
        return gauge

    def count_worker_pool_runtime(self) -> Metric:
        """Count the time spent running tasks per worker pool."""
        gauge = ZeroGauge(
            "worker_pool_runtime",
            "Execution time of WorkerPools",
            labels=("worker_pool",),
        )
        now = timezone.now()
        runtime = cast(
            Combinable, now - cast(dt.datetime, F("instance_created_at"))
        )
        # We are summing two separate joins, so do it in subqueries to
        # avoid cartesian explosion
        for stat in WorkerPool.objects.values("name").annotate(
            runtime_past=Subquery(
                WorkerPoolStatistics.objects.filter(worker_pool=OuterRef("pk"))
                .values("worker_pool")
                .annotate(sum=Sum("runtime"))
                .values("sum")
            ),
            runtime_live=Subquery(
                Worker.objects.filter(
                    worker_pool=OuterRef("pk"),
                    instance_created_at__isnull=False,
                )
                .values("worker_pool")
                .annotate(
                    sum=Sum(
                        ExpressionWrapper(
                            Extract(runtime, "epoch"),
                            output_field=IntegerField(),
                        )
                    )
                )
                .values("sum")
            ),
        ):
            gauge.add_metric(
                (stat["name"],),
                (stat["runtime_past"] or 0) + (stat["runtime_live"] or 0),
            )
        gauge.fill_zeros(self._worker_pools())
        return gauge

    def count_workflow_templates(self) -> Metric:
        """Count the number of workflow templates."""
        gauge = ZeroGauge(
            "workflow_templates",
            "Number of WorkflowTemplates",
            labels=("task_name", "scope"),
        )
        for stat in WorkflowTemplate.objects.values(
            "task_name",
            "workspace__scope__name",
        ).annotate(count=Count("id")):
            gauge.add_metric(
                (stat["task_name"], stat["workspace__scope__name"]),
                stat["count"],
            )
        gauge.fill_zeros(DBTask.task_names(TaskTypes.WORKFLOW), self._scopes())
        return gauge

    def count_workspaces(self) -> Metric:
        """Count the number of workspaces."""
        gauge = ZeroGauge(
            "workspaces",
            "Number of Workspaces",
            labels=("private", "expires", "scope"),
        )
        for stat in (
            Workspace.objects.values("public", "scope__name")
            .annotate(
                expires=ExpressionWrapper(
                    Q(expiration_delay__isnull=False),
                    output_field=BooleanField(),
                )
            )
            .annotate(count=Count("id"))
        ):
            gauge.add_metric(
                (
                    str(int(stat["public"])),
                    str(int(stat["expires"])),
                    stat["scope__name"],
                ),
                stat["count"],
            )
        gauge.fill_zeros(("0", "1"), ("0", "1"), self._scopes())
        return gauge



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""debdiff workflow."""

from collections import defaultdict
from collections.abc import Sequence
from typing import NewType, override

from debian.debian_support import Version

from debusine.artifacts import SourcePackage
from debusine.artifacts.models import (
    ArtifactCategory,
    CollectionCategory,
    DebianBinaryPackage,
    DebianSourcePackage,
    get_source_package_name,
)
from debusine.client.models import LookupChildType
from debusine.db.models import Artifact, ArtifactRelation, WorkRequest
from debusine.db.models.lookups import LookupResult
from debusine.server.collections.lookup import artifact_ids
from debusine.server.workflows import Workflow, workflow_utils
from debusine.server.workflows.models import (
    DebDiffWorkflowData,
    DebDiffWorkflowDynamicData,
    WorkRequestWorkflowData,
)
from debusine.server.workflows.workflow_utils import (
    lookup_result_architectures,
    lookup_result_binary_package_name,
)
from debusine.tasks.models import (
    DebDiffData,
    DebDiffFlags,
    DebDiffInput,
    InputArtifact,
    InputArtifactMultiple,
    InputArtifactSingle,
    LookupMultiple,
    LookupSingle,
)

PackageName = NewType("PackageName", str)
Architecture = NewType("Architecture", str)
PackageArchLookupMap = dict[PackageName, dict[Architecture, LookupSingle]]


class DebDiffWorkflow(
    Workflow[DebDiffWorkflowData, DebDiffWorkflowDynamicData]
):
    """DebDiff workflow."""

    TASK_NAME = "debdiff"

    @override
    def compute_dynamic_data(self) -> DebDiffWorkflowDynamicData:
        """
        Compute dynamic data for this workflow.

        :subject: package name of ``source_artifact``
        :parameter_summary: name of debian suite: old_version → new_version
        :source_artifact_id: id of the source artifact
        :binary_artifacts_ids: ids of the binary artifacts
        :original_id: id of the original artifact
        """
        source_artifact = workflow_utils.source_package(self)
        new_source_data = SourcePackage.create_data(source_artifact.data)
        new_source_package_name = get_source_package_name(new_source_data)
        version_new = new_source_data.version

        debian_suite = self.work_request.lookup_single(
            self.data.original, expect_type=LookupChildType.COLLECTION
        ).collection

        original_source_package = debian_suite.manager.lookup(
            f"source:{new_source_package_name}"
        )

        version_original = None
        original_source_package_id = None
        if original_source_package is not None:
            assert original_source_package.artifact is not None
            version_original = Version(
                DebianSourcePackage(
                    **original_source_package.artifact.data
                ).version
            )
            original_source_package_id = original_source_package.artifact.id
        else:
            original_binary_packages = (
                self._get_binary_packages_from_source_package(
                    debian_suite=self.data.original,
                    source_package_name=new_source_package_name,
                )
            )

            if original_binary_packages:
                for original_binary_package in original_binary_packages:
                    assert original_binary_package.artifact is not None
                    version = Version(
                        DebianBinaryPackage(
                            **original_binary_package.artifact.data
                        ).srcpkg_version
                    )
                    if version_original is None or version > version_original:
                        version_original = version

        if version_original is not None:
            parameter_summary = (
                f"{debian_suite.name}: {version_original} → {version_new}"
            )
        else:
            parameter_summary = f"{debian_suite.name}: missing → {version_new}"

        binary_artifacts_results = self.work_request.lookup_multiple(
            self.data.binary_artifacts,
            expect_type=LookupChildType.ARTIFACT_OR_PROMISE,
        )

        return DebDiffWorkflowDynamicData(
            subject=new_source_package_name,
            parameter_summary=parameter_summary,
            source_artifact_id=source_artifact.id,
            binary_artifacts_ids=artifact_ids(binary_artifacts_results),
            original_id=original_source_package_id,
        )

    @override
    def get_input_artifacts(self) -> list[InputArtifact]:
        """Return input artifacts."""
        return [
            InputArtifactSingle(
                lookup=self.data.source_artifact,
                label="source_artifact",
                artifact_id=(
                    self.dynamic_data.source_artifact_id
                    if self.dynamic_data is not None
                    else None
                ),
            ),
            InputArtifactSingle(
                lookup=self.data.original,
                label="original",
                artifact_id=(
                    self.dynamic_data.original_id
                    if self.dynamic_data is not None
                    else None
                ),
            ),
            InputArtifactMultiple.from_ids(
                lookup=self.data.binary_artifacts,
                label="binary_artifacts",
                artifact_ids=(
                    self.dynamic_data.binary_artifacts_ids
                    if self.dynamic_data is not None
                    else None
                ),
            ),
        ]

    @override
    def populate(self) -> None:
        """Create work requests."""
        new_source_data = workflow_utils.source_package_data(self)

        environment = f"{self.data.vendor}/match:codename={self.data.codename}"

        # Populate DebDiff tasks for the source package
        debian_suite = self.work_request.lookup_single(
            self.data.original, expect_type=LookupChildType.COLLECTION
        ).collection

        original_source = debian_suite.manager.lookup(
            f"source:{new_source_data.name}"
        )
        original_source_version: str | None = None
        if original_source is not None:
            original_source_version = original_source.data["version"]
            original_source_lookup = (
                f"{self.data.original}/name:{original_source.name}"
            )
            self._populate_source_artifact_debdiff(
                original=original_source_lookup,
                new=self.data.source_artifact,
                build_architecture=self.data.arch_all_build_architecture,
                environment=environment,
                extra_flags=self.data.extra_flags,
            )

        # Populate DebDiff tasks for the binary packages
        original_binaries_arch_lookup = (
            self._get_original_packages_by_name_version_arch(
                new_source_data.name, original_source_version
            )
        )
        archs_old: set[Architecture] = set()
        for archs_lookup in original_binaries_arch_lookup.values():
            archs_old.update(archs_lookup.keys())

        concrete_archs_old = archs_old - {Architecture("all")}

        new_binaries_arch_lookup = self._get_new_packages_by_name_arch()

        archs_new = self._get_architectures(new_binaries_arch_lookup)
        concrete_archs_new = archs_new - {Architecture("all")}

        if concrete_archs_new:
            # New binaries provide concrete archs: compare per concrete arch
            archs_to_compare = concrete_archs_new
        elif concrete_archs_old and Architecture("all") not in archs_old:
            # Transition any → all: new provides only "all" (because no
            # concrete_archs_new), old provides only concretes
            archs_to_compare = concrete_archs_old
        else:
            # At least one side provides "all" (and no concrete archs on new)
            archs_to_compare = {Architecture("all")}

        for arch in sorted(archs_to_compare):
            self._populate_binary_for_arch(
                architecture=arch,
                orig_map=original_binaries_arch_lookup,
                new_map=new_binaries_arch_lookup,
                environment=environment,
                extra_flags=self.data.extra_flags,
            )

    def _populate_binary_for_arch(
        self,
        *,
        architecture: Architecture,
        orig_map: PackageArchLookupMap,
        new_map: PackageArchLookupMap,
        environment: LookupSingle,
        extra_flags: list[DebDiffFlags],
    ) -> None:
        arch_all = Architecture("all")
        new_relevants = set()
        for arch_lookup in new_map.values():
            for arch in (architecture, arch_all):
                if arch in arch_lookup:
                    new_relevants.add(arch_lookup[arch])

        original_relevants: list[LookupSingle] = []
        for old_pkg_name, old_arch_map in orig_map.items():
            # Architectures are considered compatible in two cases:
            # - if comparing a concrete architecture: include the architecture
            #   and "all" packages
            # - if comparing "all": include only "all" packages
            original_relevants += [
                old_lookup
                for old_arch, old_lookup in old_arch_map.items()
                if old_pkg_name in new_map
                and (
                    (architecture == arch_all and old_arch == arch_all)
                    or (
                        architecture != arch_all
                        and old_arch in (arch_all, architecture)
                    )
                )
            ]

        if original_relevants and new_relevants:
            build_arch = (
                self.data.arch_all_build_architecture
                if architecture == Architecture("all")
                else architecture
            )

            self._populate_binary_artifact_debdiff(
                originals=LookupMultiple.model_validate(
                    sorted(original_relevants)
                ),
                news=LookupMultiple.model_validate(sorted(new_relevants)),
                architecture=architecture,
                build_architecture=build_arch,
                environment=environment,
                extra_flags=extra_flags,
            )

    def _get_binary_packages_from_source_package(
        self,
        debian_suite: LookupSingle,
        source_package_name: str,
        source_package_version: str | None = None,
    ) -> Sequence[LookupResult]:
        kwargs = {
            "collection": debian_suite,
            "child_type": LookupChildType.ARTIFACT,
            "data__srcpkg_name": source_package_name,
            "category": ArtifactCategory.BINARY_PACKAGE,
        }

        if source_package_version is not None:
            kwargs["data__srcpkg_version"] = source_package_version

        lookup = LookupMultiple.model_validate(kwargs)

        return self.work_request.lookup_multiple(
            lookup,
            expect_type=LookupChildType.ARTIFACT_OR_PROMISE,
            default_category=CollectionCategory.SUITE,
        )

    def _get_original_packages_by_name_version_arch(
        self,
        srcpkg_name: str,
        srcpkg_version: str | None,
    ) -> PackageArchLookupMap:
        """
        Return mapping of original binary packages by name and arch.

        Only packages with the highest Version. E.g. if there are the following
        packages: hello/2.0.0, libhello2/2.0.0 and libhello1/1.0.0 the last
        one will not be returned.
        """
        original_binary_results = self._get_binary_packages_from_source_package(
            debian_suite=self.data.original,
            source_package_name=srcpkg_name,
            source_package_version=srcpkg_version,
        )

        packages_by_srcpkg_version: defaultdict[
            Version, list[tuple[PackageName, Architecture, LookupSingle]]
        ] = defaultdict(list)

        for lookup_result in original_binary_results:
            architectures = lookup_result_architectures(lookup_result)
            assert len(architectures) == 1
            arch = Architecture(architectures.pop())
            pkg_name = PackageName(
                lookup_result_binary_package_name(lookup_result)
            )
            assert lookup_result.artifact is not None
            version = Version(
                DebianBinaryPackage(
                    **lookup_result.artifact.data
                ).srcpkg_version
            )

            assert lookup_result.collection_item is not None
            lookup_str = (
                f"{self.data.original}/name:"
                f"{lookup_result.collection_item.name}"
            )
            packages_by_srcpkg_version[version].append(
                (pkg_name, arch, lookup_str)
            )

        # Return packages with only the highest version
        highest_srcpkg_version = max(packages_by_srcpkg_version, default=None)
        result: PackageArchLookupMap = defaultdict(dict)
        if highest_srcpkg_version is not None:
            for pkg_name, arch, lookup in packages_by_srcpkg_version[
                highest_srcpkg_version
            ]:
                result[pkg_name][arch] = lookup
        return result

    @staticmethod
    def _extract_package_name_and_architecture(
        artifact: Artifact,
    ) -> tuple[PackageName, Architecture]:
        """Expect artifact of type BINARY_PACKAGE."""
        data = artifact.create_data()
        assert isinstance(data, DebianBinaryPackage)

        architecture = Architecture(data.deb_fields.get("Architecture", ""))

        package_name = PackageName(data.deb_fields.get("Package", ""))

        return package_name, architecture

    @classmethod
    def _package_arch_pairs(
        cls,
        lookup_result: LookupResult,
    ) -> list[tuple[PackageName, Architecture]]:
        """Return pairs of package names and architectures for a lookup."""
        if lookup_result.artifact is None:
            raise ValueError(
                "DebDiffWorkflow requires binary artifacts (not promises)"
            )

        match lookup_result.artifact.category:
            case ArtifactCategory.BINARY_PACKAGE:
                architectures = lookup_result_architectures(lookup_result)
                assert len(architectures) == 1
                architecture = Architecture(architectures.pop())
                pkg_name = PackageName(
                    lookup_result_binary_package_name(lookup_result)
                )
                return [(pkg_name, architecture)]

            case ArtifactCategory.UPLOAD:
                pairs: list[tuple[PackageName, Architecture]] = []
                for relation in lookup_result.artifact.relations.filter(
                    type=ArtifactRelation.Relations.EXTENDS,
                    target__category=ArtifactCategory.BINARY_PACKAGE,
                ).select_related("target"):
                    package_name, architecture = (
                        cls._extract_package_name_and_architecture(
                            relation.target
                        )
                    )
                    pairs.append((package_name, architecture))

                return sorted(pairs)

            # https://github.com/nedbat/coveragepy/issues/1860 (fixed in
            # coverage 7.6.2)
            case _ as artifact_category:  # pragma: no cover
                raise ValueError(f"Unexpected category {artifact_category}")

    def _get_new_packages_by_name_arch(
        self,
    ) -> PackageArchLookupMap:
        """Return mapping of new binary packages by name and arch."""
        new_binary_results = self.work_request.lookup_multiple(
            self.data.binary_artifacts,
            expect_type=LookupChildType.ARTIFACT_OR_PROMISE,
        )
        packages_by_name_arch: defaultdict[
            PackageName, dict[Architecture, LookupSingle]
        ] = defaultdict(dict)

        for result in new_binary_results:
            # DebDiffWorkflow gets populated once the build is ready
            assert result.artifact is not None

            artifact = result.artifact

            match artifact.category:
                case ArtifactCategory.BINARY_PACKAGE:
                    for pkg_name, arch in self._package_arch_pairs(result):
                        packages_by_name_arch[pkg_name][arch] = (
                            result.reconstruct(
                                workflow_root=self.work_request.workflow_root,
                            )
                        )

                case ArtifactCategory.UPLOAD:
                    for relation in artifact.relations.filter(
                        type=ArtifactRelation.Relations.EXTENDS,
                        target__category=ArtifactCategory.BINARY_PACKAGE,
                    ).select_related("target"):
                        pkg_name, arch = (
                            self._extract_package_name_and_architecture(
                                relation.target
                            )
                        )
                        packages_by_name_arch[pkg_name][arch] = (
                            relation.target.id
                        )

                # https://github.com/nedbat/coveragepy/issues/1860 (fixed in
                # coverage 7.6.2)
                case _ as artifact_category:  # pragma: no cover
                    raise ValueError(
                        f"Unexpected category {artifact_category!r}"
                    )

        return packages_by_name_arch

    def _get_architectures(
        self,
        packages_by_name_arch: dict[
            PackageName, dict[Architecture, LookupSingle]
        ],
    ) -> set[Architecture]:
        """Return all architectures present in package mapping."""
        archs: set[Architecture] = set()
        for arch_map in packages_by_name_arch.values():
            archs.update(arch_map.keys())
        return archs

    def _populate_source_artifact_debdiff(
        self,
        *,
        original: LookupSingle,
        new: LookupSingle,
        build_architecture: str,
        environment: LookupSingle,
        extra_flags: list[DebDiffFlags],
    ) -> WorkRequest:
        new = workflow_utils.locate_debian_source_package_lookup(
            self, "source_artifact", new
        )

        wr, _ = self.work_request_ensure_child_worker(
            task_name="debdiff",
            task_data=DebDiffData(
                input=DebDiffInput(
                    source_artifacts=[original, new],
                ),
                environment=environment,
                build_architecture=build_architecture,
                extra_flags=extra_flags,
            ),
            workflow_data=WorkRequestWorkflowData(
                display_name="DebDiff for source package",
                step="debdiff-source",
            ),
        )
        self.requires_artifact(wr, original)
        self.requires_artifact(wr, new)

        return wr

    def _populate_binary_artifact_debdiff(
        self,
        *,
        originals: LookupMultiple,
        news: LookupMultiple,
        architecture: str,
        build_architecture: str,
        environment: LookupSingle,
        extra_flags: list[DebDiffFlags],
    ) -> WorkRequest:
        wr, _ = self.work_request_ensure_child_worker(
            task_name="debdiff",
            task_data=DebDiffData(
                input=DebDiffInput(
                    binary_artifacts=[originals, news],
                ),
                environment=environment,
                build_architecture=build_architecture,
                extra_flags=extra_flags,
            ),
            workflow_data=WorkRequestWorkflowData(
                display_name=f"DebDiff for binary packages ({architecture})",
                step=f"debdiff-binaries-{architecture}",
            ),
        )
        self.requires_artifact(wr, originals)
        self.requires_artifact(wr, news)

        return wr



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Tests for Hetzner Cloud workers."""

import json
from typing import override
from unittest.mock import call, patch

import responses
from hcloud import APIException
from hcloud.networks.domain import Network
from responses.matchers import header_matcher, query_string_matcher

from debusine.assets import CloudProvidersType, HetznerProviderAccountData
from debusine.db.models import Asset, Worker, WorkerPool
from debusine.server.worker_pools.exceptions import InstanceLaunchException
from debusine.server.worker_pools.hetzner import HetznerCloudWorkerPool
from debusine.test.django import TestCase


class TestHetznerCloudWorkerPool(TestCase):
    """
    Test for HetznerCloudWorkerPool.

    We use responses instead of mocks, as Server() objects don't implement
    equality checking magic methods, yet:
    https://github.com/hetznercloud/hcloud-python/pull/481
    """

    provider_account: Asset
    worker_pool_model: WorkerPool

    @override
    def setUp(self) -> None:
        super().setUp()
        self.provider_account = (
            self.playground.create_cloud_provider_account_asset(
                cloud_provider=CloudProvidersType.HETZNER
            )
        )
        self.worker_pool_model = self.playground.create_worker_pool(
            provider_account=self.provider_account
        )

    def get_worker_pool(self) -> HetznerCloudWorkerPool:
        """Return a HetznerCloudWorkerPool instance."""
        worker_pool = self.worker_pool_model.provider_interface
        assert isinstance(worker_pool, HetznerCloudWorkerPool)
        return worker_pool

    @responses.activate
    def test_client_token(self) -> None:
        worker_pool = self.get_worker_pool()
        client = worker_pool.client
        self.assertIsNotNone(client)
        provider_account_data = self.provider_account.data_model
        assert isinstance(provider_account_data, HetznerProviderAccountData)
        account_token = provider_account_data.credentials.api_token
        responses.add(
            responses.GET,
            "https://api.hetzner.cloud/v1/servers",
            match=[
                header_matcher({"Authorization": f"Bearer {account_token}"})
            ],
        )
        client.request(method="GET", url="/servers")

    @responses.activate
    def test_terminate_existing_instances(self) -> None:
        worker_pool = self.get_worker_pool()
        responses.add(
            responses.DELETE,
            "https://api.hetzner.cloud/v1/servers/42",
            json={"action": {"id": 1}},
        )
        worker_pool._terminate_existing_instances({"instance-id": 42})

    def test_terminate_existing_instances_noop(self) -> None:
        worker_pool = self.get_worker_pool()
        worker_pool._terminate_existing_instances(None)
        worker_pool._terminate_existing_instances({})

    def test_terminate_existing_instances_missing(self) -> None:
        worker_pool = self.get_worker_pool()
        with patch.object(
            worker_pool.client.servers,
            "delete",
            side_effect=APIException(
                code="not_found",
                message="server with ID '42' not found",
                details=None,
            ),
        ):
            worker_pool._terminate_existing_instances({"instance-id": 42})

    def test_terminate_existing_instances_unexpected_failure(self) -> None:
        worker_pool = self.get_worker_pool()
        with (
            patch.object(
                worker_pool.client.servers,
                "delete",
                side_effect=APIException(
                    code="something_unexpected",
                    message="Server said no",
                    details=None,
                ),
            ),
            self.assertRaisesRegex(APIException, r"Server said no"),
        ):
            worker_pool._terminate_existing_instances({"instance-id": 42})

    def test_cloud_init_config(self) -> None:
        token = self.playground.create_worker_activation_token()
        config = self.get_worker_pool()._cloud_init_config("worker", token)
        self.assertIsNotNone(config)

    @responses.activate
    def test_get_network_by_name(self) -> None:
        worker_pool = self.get_worker_pool()
        responses.add(
            responses.GET,
            "https://api.hetzner.cloud/v1/networks",
            match=[query_string_matcher("name=my-network")],
            json={
                "networks": [{"id": 7, "name": "my-network"}],
            },
        )
        network = worker_pool._get_network_by_name(name="my-network")
        self.assertEqual(network.id, 7)

    @responses.activate
    def test_get_network_by_name_not_found(self) -> None:
        worker_pool = self.get_worker_pool()
        responses.add(
            responses.GET,
            "https://api.hetzner.cloud/v1/networks",
            match=[query_string_matcher("name=my-network")],
            json={
                "networks": [],
            },
        )
        with self.assertRaisesRegex(
            InstanceLaunchException, r"Unable to locate network 'my-network'"
        ):
            worker_pool._get_network_by_name(name="my-network")

    @responses.activate
    def test_create_instance_complete(self) -> None:
        token = self.playground.create_worker_activation_token()
        worker_pool = self.get_worker_pool()
        worker_pool.specifications.image_name = "debian-13"
        worker_pool.specifications.ssh_keys = ["testkey"]
        worker_pool.specifications.location = "nbg1"
        worker_pool.specifications.networks = ["my-net"]
        worker_pool.specifications.labels = {"role": "debusine-worker"}
        worker_pool.specifications.enable_ipv4 = True
        worker_pool.specifications.enable_ipv6 = True
        response = responses.add(
            responses.POST,
            "https://api.hetzner.cloud/v1/servers",
            json={
                "action": {"id": 1},
                "server": {"id": 42},
                "next_actions": [],
                "root_password": "secret-password",
            },
        )
        with patch.object(
            worker_pool, "_get_network_by_name", return_value=Network(id=7)
        ):
            server_id = worker_pool._create_instance(
                worker_name="hetzner-001", activation_token=token
            )
        self.assertEqual(server_id, 42)

        assert response.calls[0].request.body
        body = json.loads(response.calls[0].request.body)
        self.assertEqual(body["name"], "hetzner-001")
        self.assertEqual(body["image"], worker_pool.specifications.image_name)
        self.assertEqual(body["ssh_keys"], worker_pool.specifications.ssh_keys)
        self.assertEqual(body["location"], worker_pool.specifications.location)
        self.assertEqual(body["networks"], [7])
        self.assertTrue(body["user_data"].startswith("#cloud-config"))
        self.assertEqual(body["labels"], worker_pool.specifications.labels)
        self.assertEqual(
            body["public_net"]["enable_ipv4"],
            worker_pool.specifications.enable_ipv4,
        )
        self.assertEqual(
            body["public_net"]["enable_ipv6"],
            worker_pool.specifications.enable_ipv6,
        )

    @responses.activate
    def test_create_instance_simple(self) -> None:
        token = self.playground.create_worker_activation_token()
        worker_pool = self.get_worker_pool()
        worker_pool.specifications.image_name = "debian-13"
        worker_pool.specifications.ssh_keys = []
        worker_pool.specifications.location = None
        worker_pool.specifications.networks = []
        worker_pool.specifications.labels = {}
        worker_pool.specifications.enable_ipv4 = False
        worker_pool.specifications.enable_ipv6 = False
        response = responses.add(
            responses.POST,
            "https://api.hetzner.cloud/v1/servers",
            json={
                "action": {"id": 1},
                "server": {"id": 42},
                "next_actions": [],
                "root_password": "secret-password",
            },
        )
        server_id = worker_pool._create_instance(
            worker_name="hetzner-001", activation_token=token
        )
        self.assertEqual(server_id, 42)

        assert response.calls[0].request.body
        body = json.loads(response.calls[0].request.body)
        self.assertEqual(body["name"], "hetzner-001")
        self.assertEqual(body["image"], worker_pool.specifications.image_name)
        self.assertEqual(body["ssh_keys"], worker_pool.specifications.ssh_keys)
        self.assertNotIn("location", body)
        self.assertEqual(body["networks"], [])
        self.assertTrue(body["user_data"].startswith("#cloud-config"))
        self.assertEqual(body["labels"], worker_pool.specifications.labels)
        self.assertEqual(
            body["public_net"]["enable_ipv4"],
            worker_pool.specifications.enable_ipv4,
        )
        self.assertEqual(
            body["public_net"]["enable_ipv6"],
            worker_pool.specifications.enable_ipv6,
        )

    def test_launch_worker(self) -> None:
        worker_pool = self.get_worker_pool()
        worker = self.playground.create_worker(
            worker_pool=worker_pool.worker_pool
        )
        worker.worker_pool_data = {"random-data": "foo"}
        worker.instance_created_at = None
        worker.save()

        with (
            patch.object(
                worker_pool, "_create_instance", return_value=42
            ) as create_instance,
            patch.object(
                worker_pool, "_terminate_existing_instances"
            ) as terminate_existing_instances,
        ):
            worker_pool.launch_worker(worker)

        create_instance.assert_called_once_with(
            worker_name=worker.name,
            activation_token=worker.activation_token,
        )
        terminate_existing_instances.assert_called_once_with(
            {"random-data": "foo"}
        )
        self.assertIsNotNone(worker.instance_created_at)
        self.assertEqual(worker.worker_pool_data, {"instance-id": 42})
        assert worker.activation_token
        self.assertTrue(worker.activation_token.enabled)

    def test_terminate_worker(self) -> None:
        worker_pool = self.get_worker_pool()
        worker = self.playground.create_worker(
            worker_pool=worker_pool.worker_pool
        )
        worker.worker_pool_data = {"instance-id": 42}
        worker.save()

        with patch.object(
            worker_pool, "_terminate_existing_instances"
        ) as terminate_existing_instances:
            worker_pool.terminate_worker(worker)

        terminate_existing_instances.assert_called_once_with(
            {"instance-id": 42}
        )
        self.assertIsNone(worker.instance_created_at)
        self.assertIsNone(worker.worker_pool_data)

    @responses.activate
    def test_terminate_missing(self) -> None:
        worker_pool = self.get_worker_pool()
        workers: dict[int, Worker] = {}
        for instance_id in (11, 12, 13):
            worker = self.playground.create_worker(
                worker_pool=worker_pool.worker_pool
            )
            worker.worker_pool_data = {"instance-id": instance_id}
            worker.save()
            workers[instance_id] = worker

        task = self.playground.create_worker_task(
            mark_running=True, worker=workers[13]
        )

        # Inactive pool member
        worker = self.playground.create_worker(
            worker_pool=worker_pool.worker_pool
        )
        worker.instance_created_at = None
        worker.save()

        responses.add(
            responses.GET,
            "https://api.hetzner.cloud/v1/servers",
            json={
                "servers": [
                    {
                        "id": 11,
                        "status": "running",
                    },
                    {
                        "id": 12,
                        "status": "off",
                    },
                ]
            },
        )
        with (
            patch.object(
                worker_pool, "_terminate_existing_instances"
            ) as terminate_existing_instances,
        ):
            worker_pool.terminate_missing()

        self.assertEqual(terminate_existing_instances.call_count, 2)
        terminate_existing_instances.assert_has_calls(
            [
                call({"instance-id": 12}),
                call({"instance-id": 13}),
            ],
            any_order=True,
        )
        task.refresh_from_db()
        self.assertIsNone(task.worker)

    @responses.activate
    def test_terminate_missing_empty(self) -> None:
        worker_pool = self.get_worker_pool()
        worker_pool.terminate_missing()
        self.assertEqual(len(responses.calls), 0)
